"""Provide the SpikingJellyMultiStep learningframework module."""

import sys

import torch
from spikingjelly.activation_based import functional  # type: ignore[import-untyped]

from .SpikingJelly import SpikingJelly

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class SpikingJellyMultiStep(SpikingJelly):
    """SpikingJelly multi-step LearningFramework implementation extending SpikingJelly single-step."""

    class TrainerModule(SpikingJelly.TrainerModule):
        """SpikingJelly multi-step TrainerModule implementation extending SpikingJelly single-step TrainerModule."""

        @override
        def forward(self, x: torch.Tensor) -> torch.Tensor:
            """Forward pass for a Spiking Neural Network model with duplicated timesteps in multi-step mode.

            First calls SpikingJelly's reset on the model to reset neurons potentials.
            Then duplicate the input to generate the number of timesteps given by
            :attr:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.timesteps`.
            Timesteps are generated as a new dimension of the input tensor: [N, C, …] → [T, N, C, …]
            Call :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.forward` for the whole tensor.
            Finally, average the output of the model over the timestep dimension.

            :param x: Input data
            :return: Output predictions
            """
            functional.reset_net(self.model)

            # Duplicate input over timesteps
            x = x.unsqueeze(0).repeat(self.model.timesteps, *([1] * len(x.shape)))

            return self.model(x).sum(0) / self.model.timesteps



"""ATIF-u: spiking neuron with learnable quantization steps.

Author: Andrea Castagetti <Andrea.CASTAGNETTI@univ-cotedazur.fr>
"""

from __future__ import annotations

import logging
import sys
from typing import Any, Literal, cast

import torch
from qualia_core.typing import TYPE_CHECKING
from spikingjelly.activation_based.neuron import BaseNode  # type: ignore[import-untyped]
from torch import nn

if TYPE_CHECKING:
    from torch.autograd.function import FunctionCtx  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

def heaviside(x: torch.Tensor) -> torch.Tensor:
    """Heaviside function.

    :param x: Input tensor
    :return: Boolean tensor of the same dimension as `x` where each element is ``True`` if the element in `x` is greater than or
             equal to 0, ``False`` otherwise
    """
    return (x >= 0).to(x)

class SpikeFunctionSigmoid(torch.autograd.Function):
    """Spike functions with surrogate bkw gradient."""

    @staticmethod
    @override
    def forward(ctx: FunctionCtx, *args: torch.Tensor, **_: Any) -> torch.Tensor:
        """Forward of :func:`heaviside` function.

        :param ctx: A context object used to save tensors for :meth:`backward`
        :param args: Tuple of 2 tensors for ``x`` and ``alpha``, respectively, saved in ``ctx`` for backward pass
        :param _: Unused
        :return: Tensor of :func:`heaviside` applied over ``x``.
        """
        x, alpha = args
        if x.requires_grad:
            ctx.save_for_backward(x, alpha)
        return heaviside(x)

    @staticmethod
    @override
    def backward(ctx: torch.autograd.Function, *grad_outputs: torch.Tensor) -> tuple[torch.Tensor | None, None]:
        """Backward pass of surrogate gradient using :meth:`torch.Tensor.sigmoid_` function.

        :param ctx: Context to restore the ``x`` and ``alpha`` tensors from
        :param grad_outputs: Output tensor from the computation of the :meth:`forward` pass
        :return: A tuple of Tensor and None with the first element being the computed gradient for ``x`` or None if there is no
                 gradient to compute and the second element a placeholder for the gradient of ``alpha``.
        """
        grad_x = None
        grad_output = grad_outputs[0]
        if ctx.needs_input_grad[0]:
            x, alpha = cast(tuple[torch.Tensor, torch.Tensor], ctx.saved_tensors) # Couple of tensors saved in forward()
            device = x.device
            sgax = ((x * alpha.to(device=device)).sigmoid_()).to(device=device)
            grad_x = (grad_output.to(device=device)
                      * (1. - sgax.to(device=device))
                      * sgax.to(device=device)
                      * alpha.to(device=device))
        return grad_x, None

    @classmethod
    @override
    def apply(cls, *args: torch.Tensor, **kwargs: Any) -> torch.Tensor:
        """Apply heaviside activation with sigmoid surrogate gradient.

        :param args: Input tensor
        :param kwargs: Unused
        :return: Output tensor
        """
        return cast(torch.Tensor, super().apply(*args, **kwargs))  # type: ignore[no-untyped-call]

class IFSRL(nn.Module):
    """IFSRL: Integrate and Fire soft-reset with learnable Vth and activation scaling."""

    v: torch.Tensor

    def __init__(self,  # noqa: PLR0913
                 v_threshold: float = 1.0,
                 vth_init_l: float = 0.8,
                 vth_init_h: float = 1.,
                 alpha: float = 1.,
                 device: str = 'cpu') -> None:
        """Construct :class:`IFSRL`.

        :param v_threshold: Factor to apply to the uniform initialization bounds
        :param vth_init_l: Lower bound for uniform initialization of threshold Tensor
        :param vth_init_h: Higher bound for uniform initialization of threshold Tensor
        :param alpha: Sigmoig surrogate scale factor
        :param device: Device to run the computation on
        """
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__()
        self.v_threshold = v_threshold
        self.alpha = alpha
        self.device = device

        self.vp_th = torch.nn.Parameter(torch.ones(1).to(self.device), requires_grad=True)
        self.v = torch.zeros(1, device=device)

        with torch.no_grad():
            _ = nn.init.uniform_(self.vp_th,
                                 a=vth_init_l * self.v_threshold,
                                 b=vth_init_h * self.v_threshold)

    def get_coeffs(self) -> torch.Tensor:
        """Return the Tensor of threshold :attr:`vp_th`.

        :return: Tensor of threshold :attr:`vp_th`
        """
        return self.vp_th

    def set_coeffs(self, vp_th: torch.Tensor) -> None:
        """Replace the Tensor of threshold :attr:`vp_th`.

        :param vp_th: New Tensor of threshold to replace :attr:`vp_th`
        """
        _ = self.vp_th.copy_(vp_th)

    def reset(self) -> None:
        """Reset potential to 0."""
        _ = self.v.zero_() # midrise quantizer

    def ifsrl_fn(self, x: torch.Tensor) -> torch.Tensor:
        """Integrate-and-Fire soft-reset neuron with learnable threshold.

        :param x: Input tensor
        :return: Output tensor
        """
        # Primary membrane charge
        self.v = self.v + x

        # Fire
        q = (self.v - self.vp_th)
        z = SpikeFunctionSigmoid.apply(q, self.alpha * torch.ones(1).to(self.device)).float()

        # Soft-Reset
        self.v = (1. - z) * self.v + z * (self.v - self.vp_th)

        return z

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward of :meth:`ifsrl_fn`.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input
        return self.ifsrl_fn(x)


class ATIF(BaseNode):  # type: ignore[misc]
    """IFSRLSJ: Integrate and Fire soft-reset with learnable Vth and activation scaling, based on spikingjelly."""

    v: torch.Tensor

    def __init__(self,  # noqa: PLR0913
                 v_threshold: float = 1.0,
                 vth_init_l: float = 0.8,
                 vth_init_h: float = 1.,
                 alpha: float = 1.,
                 device: str = 'cpu') -> None:
        """Construct :class:`ATIF`.

        :param v_threshold: Factor to apply to the uniform initialization bounds
        :param vth_init_l: Lower bound for uniform initialization of threshold Tensor
        :param vth_init_h: Higher bound for uniform initialization of threshold Tensor
        :param alpha: Sigmoig surrogate scale factor
        :param device: Device to run the computation on
        """
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__(v_threshold=v_threshold,
                         v_reset=None,
                         surrogate_function=SpikeFunctionSigmoid.apply,
                         step_mode='s') # init the base class
        self.device = device
        self.alpha = alpha
        self.v_threshold = torch.nn.Parameter(torch.ones(1).to(self.device) * v_threshold, requires_grad=True)

        self.v = torch.zeros(1, device=device)

        with torch.no_grad():
            _= nn.init.uniform_(self.v_threshold,
                             a=vth_init_l * v_threshold,
                             b=vth_init_h * v_threshold)

    @property
    @override  # type: ignore[misc]
    def supported_backends(self) -> tuple[Literal['torch']]:
        """Supported step_mode and backend.

        Only single-step mode with torch backend is supported.

        :return: Tuple of ``'torch'`` if :attr:`step_mode` is ``'s'``
        :raise ValueError: When :attr:`step_mode` is not ``'s'``
        """
        if self.step_mode == 's':
            return ('torch',)
        logger.error("Only step_mode='s' is supported, current step_mode='%s'", self.step_mode)
        raise ValueError

    def get_coeffs(self) -> torch.Tensor:
        """Return the Tensor of threshold :attr:`v_threshold`.

        :return: Tensor of threshold :attr:`v_threshold`
        """
        return self.v_threshold

    def set_coeffs(self, v_threshold: torch.Tensor) -> None:
        """Replace the Tensor of threshold :attr:`v_threshold`.

        :param v_threshold: New Tensor of threshold to replace :attr:`v_threshold`
        """
        _ = self.v_threshold.copy_(v_threshold)

    def ifsrl_fn(self, x: torch.Tensor) -> torch.Tensor:
        """Integrate-and-Fire soft-reset neuron with learnable threshold.

        :param x: Input tensor
        :return: Output tensor
        """
        # Primary membrane charge
        self.v_float_to_tensor(x)
        self.v = self.v + x

        # Fire
        q = (self.v - self.v_threshold)
        z = SpikeFunctionSigmoid.apply(q, self.alpha * torch.ones(1).to(self.device)).float()

        # Soft-Reset
        self.v = (1. - z) * self.v + z * (self.v - self.v_threshold)

        return z * self.get_coeffs()

    @override  # type: ignore[misc]
    def single_step_forward(self, x: torch.Tensor) -> torch.Tensor:
        """Single-step mode forward of ATIF.

        Calls :meth:`ifsrl_fn`.

        :param x: Input tensor
        :return: Output tensor
        """
        return self.ifsrl_fn(x)



"""Contains the template for a residual spiking neural network."""

from __future__ import annotations

import logging
import sys
from typing import Callable, Protocol, cast

import numpy as np
import torch
from qualia_core.learningmodel.pytorch.layers import Add
from qualia_core.typing import TYPE_CHECKING
from torch import nn

from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly import layers1d as sjlayers1d
from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly import layers2d as sjlayers2d

from .SNN import SNN

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    from qualia_core.typing import RecursiveConfigDict

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class BasicBlockBuilder(Protocol):
    """Signature for basicblockbuilder.

    Used to bind hyperparameters constant across all the ResNet blocks.
    """

    def __call__(self,  # noqa: PLR0913
                 in_planes: int,
                 planes: int,
                 kernel_size: int,
                 stride: int,
                 padding: int) -> BasicBlock:
        """Build a :class:`BasicBlock`.

        :param in_planes: Number of input channels
        :param planes: Number of filters (i.e., output channels) in the main branch Conv layers
        :param kernel_size: ``kernel_size`` for the main branch Conv layers
        :param stride: ``kernel_size`` for the MaxPool layers, no MaxPool layer added if `1`
        :param padding: Padding for the main branch Conv layers
        :return: A :class:`BasicBlock`
        """
        ...

class BasicBlock(nn.Module):
    r"""A single ResNetv1 block.

    Structure is:

    .. code-block::

                |
              /   \
            |       |
           Conv     |
            |       |
        BatchNorm   |
            |       |
         MaxPool   Conv
            |       |
            IF  BatchNorm
            |       |
           Conv  MaxPool
            |       |
        BatchNorm   IF
            |       |
            IF      |
            |       |
              \   /
                |
               Add

    Main (left) branch Conv use ``kernel_size=kernel_size``, while residual (right) branch Conv use ``kernel_size=1``.

    BatchNorm layers will be absent if ``batch_norm == False``

    MaxPool layer will be absent if ``stride == 1``.

    Residual (right) branch Conv layer will be asbent if ``in_planes==planes``, except if ``force_projection_with_stride==True``
    and ``stride != 1``.
    """

    expansion: int = 1 #: Unused

    def __init__(self,  # noqa: PLR0913
                 sjlayers_t: ModuleType,
                 in_planes: int,
                 planes: int,
                 kernel_size: int,
                 stride: int,
                 padding: int,
                 batch_norm: bool,  # noqa: FBT001
                 bn_momentum: float,
                 force_projection_with_stride: bool,  # noqa: FBT001
                 create_neuron: Callable[[], nn.Module],
                 step_mode: str) -> None:
        """Construct :class:`BasicBlock`.

        :param sjlayers_t: Module containing the aliased layers to use (1D or 2D)
        :param in_planes: Number of input channels
        :param planes: Number of filters (i.e., output channels) in the main branch Conv layers
        :param kernel_size: ``kernel_size`` for the main branch Conv layers
        :param stride: ``kernel_size`` for the MaxPool layers, no MaxPool layer added if `1`
        :param padding: Padding for the main branch Conv layers
        :param batch_norm: If ``True``, add BatchNorm layer after each Conv layer
        :param bn_momentum: BatchNorm layer ``momentum``
        :param force_projection_with_stride: If ``True``, residual Conv layer is kept when ``stride != 1`` even if
                                             ``in_planes == planes``
        :param create_neuron: :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.create_neuron` method to instantiate a spiking
                              neuron
        :param step_mode: SpikingJelly ``step_mode`` from :attr:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.step_mode`
        """
        super().__init__()
        self.expansion = 1
        self.batch_norm = batch_norm
        self.force_projection_with_stride = force_projection_with_stride
        self.in_planes = in_planes
        self.planes = planes
        self.stride = stride

        self.conv1 = sjlayers_t.Conv(in_planes,
                                     planes,
                                     kernel_size=kernel_size,
                                     stride=1,
                                     padding=padding,
                                     bias=not batch_norm,
                                     step_mode=step_mode)
        if batch_norm:
            self.bn1 = sjlayers_t.BatchNorm(planes, momentum=bn_momentum, step_mode=step_mode)

        if self.stride != 1:
            self.pool1 = sjlayers_t.MaxPool(stride, step_mode=step_mode)

        # QUANTIZED NEURON
        self.neuron1 = create_neuron()

        self.conv2 = sjlayers_t.Conv(planes,
                                     planes,
                                     kernel_size=kernel_size,
                                     stride=1,
                                     padding=padding,
                                     bias=not batch_norm,
                                     step_mode=step_mode)

        if batch_norm:
            self.bn2 = sjlayers_t.BatchNorm(planes, momentum=bn_momentum, step_mode=step_mode)

        self.neuron2 = create_neuron()
        # residual path
        if self.stride != 1:
            self.smax = sjlayers_t.MaxPool(stride, step_mode=step_mode)
        if self.in_planes != self.expansion*self.planes or force_projection_with_stride and self.stride != 1:
            self.sconv = sjlayers_t.Conv(in_planes,
                                         self.expansion*planes,
                                         kernel_size=1,
                                         stride=1,
                                         bias=not batch_norm,
                                         step_mode=step_mode)
            self.neuronr = create_neuron()
            if batch_norm:
                self.sbn = sjlayers_t.BatchNorm(self.expansion*planes, momentum=bn_momentum, step_mode=step_mode)
        self.add = Add()

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward of ResNet block.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input

        out = self.conv1(x)

        if self.batch_norm:
            out = self.bn1(out)
        if self.stride != 1:
            out = self.pool1(out)
        out = self.neuron1(out)

        out = self.conv2(out)
        if self.batch_norm:
            out = self.bn2(out)
        out = self.neuron2(out)

        # shortcut
        tmp = x

        if self.in_planes != self.expansion*self.planes or self.force_projection_with_stride and self.stride != 1:
            tmp = self.sconv(tmp)

            if self.batch_norm:
                tmp = self.sbn(tmp)

        if self.stride != 1:
            tmp = self.smax(tmp)

        if self.in_planes != self.expansion*self.planes or self.force_projection_with_stride and self.stride != 1:
            tmp = self.neuronr(tmp)

        return self.add(out, tmp)

class SResNet(SNN):
    """Residual spiking neural network template.

    Similar to :class:`qualia_core.learningmodel.pytorch.ResNet.ResNet` but with spiking neuron activation layers (e.g., IF)
    instead of :class:`torch.nn.ReLU`.


    Example TOML configuration for a 2D ResNetv1-18 over 4 timesteps with soft-reset multi-step IF based on the SResNet template:

    .. code-block:: toml

        [[model]]
        name = "SResNetv1-18"
        params.filters      = [64, 64, 128, 256, 512]
        params.kernel_sizes = [ 7,  3,   3,   3,   3]
        params.paddings     = [ 3,  1,   1,   1,   1]
        params.strides      = [ 2,  1,   1,   1,   1]
        params.num_blocks   = [     2,   2,   2,   2]
        params.prepool      = 1
        params.postpool     = 'max'
        params.batch_norm   = true
        params.dims         = 2
        params.timesteps    = 4
        params.neuron.kind  = 'IFNode'
        params.neuron.params.v_reset = false # Soft reset
        params.neuron.params.v_threshold = 1.0
        params.neuron.params.detach_reset = true
        params.neuron.params.step_mode = 'm' # Multi-step mode, make sure to use SpikingJellyMultiStep learningframework
        params.neuron.params.backend = 'torch'
    """

    def __init__(self,  # noqa: PLR0913, C901
                 input_shape: tuple[int, ...],
                 output_shape: tuple[int, ...],
                 filters: list[int],
                 kernel_sizes: list[int],

                 num_blocks: list[int],
                 strides: list[int],
                 paddings: list[int],
                 prepool: int = 1,
                 postpool: str = 'max',
                 batch_norm: bool = False,  # noqa: FBT001, FBT002
                 bn_momentum: float = 0.1,
                 force_projection_with_stride: bool = True,  # noqa: FBT001, FBT002

                 neuron: RecursiveConfigDict | None = None,
                 timesteps: int = 2,

                 dims: int = 1,
                 basicblockbuilder: BasicBlockBuilder | None = None) -> None:
        """Construct :class:`SResNet`.

        Structure is:

        .. code-block::

              Input
                |
             AvgPool
                |
               Conv
                |
            BatchNorm
                |
                IF
                |
            BasicBlock
                |
                …
                |
            BasicBlock
                |
            GlobalPool
                |
             Flatten
                |
              Linear

        :param input_shape: Input shape
        :param output_shape: Output shape
        :param filters: List of ``out_channels`` for Conv layers inside each :class:`BasicBlock` group, must be of the same size
                        as ``num_blocks``, first element is for the first Conv layer at the beginning of the network
        :param kernel_sizes: List of ``kernel_size`` for Conv layers inside each :class:`BasicBlock` group, must of the same size
                             as ``num_blocks``, first element is for the first Conv layer at the beginning of the network
        :param num_blocks: List of number of :class:`BasicBlock` in each group, also defines the number of :class:`BasicBlock`
                           groups inside the network
        :param strides: List of ``kernel_size`` for MaxPool layers inside each :class:`BasicBlock` group, must of the same size as
                        ``num_blocks``, ``stride`` is applied only to the first :class:`BasicBlock` of the group, next
                        :class:`BasicBlock` in the group use a ``stride`` of ``1``, first element is the stride of the first Conv
                        layer at the beginning of the network
        :param paddings: List of ``padding`` for Conv layer inside each :class:`BasicBlock` group, must of the same size
                         as ``num_blocks``, first element is for the first Conv layer at the beginning of the network
        :param prepool: AvgPool layer ``kernel_size`` to add at the beginning of the network, no layer added if 0
        :param postpool: Global pooling layer type after all :class:`BasicBlock`, either `max` for MaxPool or `avg` for AvgPool
        :param batch_norm: If ``True``, add a BatchNorm layer after each Conv layer, otherwise no layer added
        :param bn_momentum: BatchNorm ``momentum``
        :param force_projection_with_stride: If ``True``, residual Conv layer is kept when ``stride != 1`` even if
                                             ``in_planes == planes`` inside a :class:`BasicBlock`
        :param neuron: Spiking neuron configuration, see :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.__init__`
        :param timesteps: Number of timesteps
        :param dims: Either 1 or 2 for 1D or 2D convolutional network.
        :param basicblockbuilder: Optional function with :meth:`BasicBlockBuilder.__call__` signature to build a basic block after
                                  binding constants common across all basic blocks
        """
        super().__init__(input_shape=input_shape,
                         output_shape=output_shape,
                         timesteps=timesteps,
                         neuron=neuron)

        from spikingjelly.activation_based.layer import Flatten, Linear  # type: ignore[import-untyped]

        sjlayers_t: ModuleType
        if dims == 1:
            sjlayers_t = sjlayers1d
        elif dims == 2:  # noqa: PLR2004
            sjlayers_t = sjlayers2d
        else:
            logger.error('Only dims=1 or dims=2 supported, got: %s', dims)
            raise ValueError

        if basicblockbuilder is None:
            def builder(in_planes: int,
                        planes: int,
                        kernel_size: int,
                        stride: int,
                        padding: int) -> BasicBlock:
                return BasicBlock(
                        sjlayers_t=sjlayers_t,
                        in_planes=in_planes,
                        planes=planes,
                        kernel_size=kernel_size,
                        stride=stride,
                        padding=padding,
                        batch_norm=batch_norm,
                        bn_momentum=bn_momentum,
                        force_projection_with_stride=force_projection_with_stride,
                        create_neuron=self.create_neuron,
                        step_mode=self.step_mode)
            basicblockbuilder = builder

        self.in_planes: int = filters[0]
        self.batch_norm = batch_norm
        self.num_blocks = num_blocks

        if prepool > 1:
            self.prepool = sjlayers_t.AvgPool(prepool, step_mode=self.step_mode)

        self.conv1 = sjlayers_t.Conv(input_shape[-1],
                                     filters[0],
                                     kernel_size=kernel_sizes[0],
                                     stride=strides[0],
                                     padding=paddings[0],
                                     bias=not batch_norm,
                                     step_mode=self.step_mode)
        if self.batch_norm:
            self.bn1 = sjlayers_t.BatchNorm(self.in_planes,
                                            momentum=bn_momentum,
                                            step_mode=self.step_mode)

        self.neuron1 = self.create_neuron()

        blocks: list[nn.Sequential] = []
        for planes, kernel_size, stride, padding, num_block in zip(filters[1:],
                                                                   kernel_sizes[1:],
                                                                   strides[1:],
                                                                   paddings[1:],
                                                                   num_blocks):
            blocks.append(self._make_layer(basicblockbuilder, num_block, planes, kernel_size, stride, padding))
        self.layers = nn.ModuleList(blocks)

        # GlobalMaxPool kernel_size computation
        self._fm_dims = np.array(input_shape[:-1]) // np.array(prepool)
        for _, kernel, stride, padding in zip(filters, kernel_sizes, strides, paddings):
            self._fm_dims += np.array(padding) * 2
            self._fm_dims -= (kernel - 1)
            self._fm_dims = np.floor(self._fm_dims / stride).astype(int)

        if postpool == 'avg':
            self.postpool = sjlayers_t.AvgPool(tuple(self._fm_dims), step_mode=self.step_mode)
        elif postpool == 'max':
            self.postpool = sjlayers_t.MaxPool(tuple(self._fm_dims), step_mode=self.step_mode)

        self.flatten = Flatten(step_mode=self.step_mode)
        self.linear = Linear(self.in_planes*BasicBlock.expansion, output_shape[0], step_mode=self.step_mode)

    def _make_layer(self,  # noqa: PLR0913
                    basicblockbuilder: BasicBlockBuilder,
                    num_blocks: int,
                    planes: int,
                    kernel_size: int,
                    stride: int,
                    padding: int) -> nn.Sequential:
        strides = [stride] + [1]*(num_blocks-1)
        layers: list[BasicBlock] = []
        for stride in strides:
            block: BasicBlock = basicblockbuilder(in_planes=self.in_planes,
                                      planes=planes,
                                      kernel_size=kernel_size,
                                      stride=stride,
                                      padding=padding)
            layers.append(block)
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward of residual spiking neural network.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input
        if hasattr(self, 'prepool'):
            x = self.prepool(x)

        out = self.conv1(x)
        if self.batch_norm:
            out = self.bn1(out)
        out = self.neuron1(out)

        for i in range(len(self.layers)):
            out = self.layers[i](out)

        if hasattr(self, 'postpool'):
            out = self.postpool(out)

        out = self.flatten(out)
        return cast(torch.Tensor, self.linear(out))



"""Contains the template for a quantized convolutional spiking neural network."""

from __future__ import annotations

import logging
import math
import sys
from collections import OrderedDict

import numpy as np
from qualia_core.learningmodel.pytorch import layers1d, layers2d
from qualia_core.learningmodel.pytorch.layers.quantized_layers import QuantizedIdentity
from qualia_core.typing import TYPE_CHECKING
from torch import nn

from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly import layers1d as sjlayers1d
from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly import layers2d as sjlayers2d

from .SNN import SNN

if TYPE_CHECKING:
    from types import ModuleType  # noqa: TCH003

    import torch
    from qualia_core.learningmodel.pytorch.Quantizer import QuantizationConfig  # noqa: TCH002
    from qualia_core.typing import RecursiveConfigDict

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

class QuantizedSCNN(SNN):
    """Quantized convolutional spiking neural network template.

    Should have topology identical to :class:`qualia_plugin_snn.learningmodel.pytorch.SCNN.SCNN` but with layers replaced with
    their quantized equivalent.
    """

    layers: nn.ModuleDict #: List of sequential layers of the SCNN model


    def __init__(self,  # noqa: PLR0913, PLR0915, PLR0912, C901
                 input_shape: tuple[int, ...],
                 output_shape: tuple[int, ...],

                 filters: list[int],
                 kernel_sizes: list[int],
                 paddings: list[int],
                 strides: list[int],
                 dropouts: float | list[float],
                 pool_sizes: list[int],
                 fc_units: list[int],

                 quant_params: QuantizationConfig,

                 batch_norm: bool = False,  # noqa: FBT001, FBT002
                 prepool: int | list[int] = 1,
                 postpool: int | list[int] = 1,

                 neuron: RecursiveConfigDict | None = None,
                 timesteps: int = 4,

                 gsp: bool = False,  # noqa: FBT001, FBT002

                 dims: int=1) -> None:
        """Construct :class:`QuantizedSCNN`.

        :param input_shape: Input shape
        :param output_shape: Output shape
        :param filters: List of ``out_channels`` for each QuantizedConv layer, also defines the number of QuantizedConv layers
        :param kernel_sizes: List of ``kernel_size`` for each QuantizedConv layer, must of the same size as ``filters``
        :param paddings: List of ``padding`` for each QuantizedConv layer, must of the same size as ``filters``
        :param strides: List of ``stride`` for each QuantizedConv layer, must of the same size as ``filters``
        :param dropouts: List of Dropout layer ``p`` to apply after each QuantizedConv or QuantizedLinear layer, must be of the
                         same size as ``filters`` + ``fc_units``, no layer added if element is 0
        :param pool_sizes: List of QuantizedMaxPool layer ``kernel_size`` to apply after each QuantizedConv layer, must be of the
                           same size as ``filters``, no layer added if element is 0
        :param fc_units: List of QuantizedLinear layer ``out_features`` to add at the end of the network, no layer added
                         if empty
        :param batch_norm: If ``True``, add a QuantizedBatchNorm layer after each QuantizedConv layer, otherwise no layer added
        :param prepool: QuantizedAvgPool layer ``kernel_size`` to add at the beginning of the network, no layer added if 0
        :param postpool: QuantizedAvgPool layer ``kernel_size`` to add after all QuantizedConv layers, no layer added if 0
        :param neuron: Spiking neuron configuration, see :meth:`qualia_plugin_snn.learningmodel.pytorch.SNN.SNN.__init__`,
                       ``'Quantized'`` is automatically prepended to ``neuron.kind``
        :param timesteps: Number of timesteps
        :param gsp: If ``True``, a single QuantizedGlobalSumPool layer is added instead of QuantizedLinear layers
        :param dims: Either 1 or 2 for 1D or 2D convolutional network.
        :param quant_params: Quantization configuration dict, see :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        """
        # Prepend Quantized to the neuron kind to instantiate quantized spiking neurons
        if neuron is not None and 'kind' in neuron and isinstance(neuron['kind'], str):
            neuron['kind'] = 'Quantized' + neuron['kind']

        super().__init__(input_shape=input_shape,
                         output_shape=output_shape,
                         timesteps=timesteps,
                         neuron=neuron)

        from spikingjelly.activation_based.layer import Dropout, Flatten  # type: ignore[import-untyped]

        from qualia_plugin_snn.learningmodel.pytorch.layers.spikingjelly.quantized_layers import QuantizedLinear

        layers_t: ModuleType
        sjlayers_t: ModuleType

        if dims == 1:
            layers_t = layers1d
            sjlayers_t = sjlayers1d
        elif dims == 2:  # noqa: PLR2004
            layers_t = layers2d
            sjlayers_t = sjlayers2d
        else:
            logger.error('Only dims=1 or dims=2 supported, got: %s', dims)
            raise ValueError

        # Backward compatibility for config not defining dropout as a list
        dropout_list = [dropouts] * (len(filters) + len(fc_units)) if not isinstance(dropouts, list) else dropouts

        layers: OrderedDict[str, nn.Module] = OrderedDict()

        layers['identity1'] = QuantizedIdentity(quant_params=quant_params)

        if (math.prod(prepool) if isinstance(prepool, list) else prepool) > 1:
            layers['prepool'] = sjlayers_t.QuantizedAvgPool(tuple(prepool) if isinstance(prepool, list) else prepool,
                                                   step_mode=self.step_mode,
                                                   quant_params=quant_params)

        layers['conv1'] = sjlayers_t.QuantizedConv(in_channels=input_shape[-1],
                                                   out_channels=filters[0],
                                                   kernel_size=kernel_sizes[0],
                                                   padding=paddings[0],
                                                   stride=strides[0],
                                                   bias=not batch_norm,
                                                   quant_params=quant_params,
                                                   step_mode=self.step_mode)

        if batch_norm:
            layers['bn1'] = sjlayers_t.QuantizedBatchNorm(filters[0], quant_params=quant_params, step_mode=self.step_mode)

        layers['neuron1'] = self.create_neuron(quant_params=quant_params)

        if dropout_list[0]:
            layers['dropout1'] = Dropout(dropout_list[0], step_mode=self.step_mode)
        if pool_sizes[0]:
            layers['maxpool1'] = sjlayers_t.QuantizedMaxPool(pool_sizes[0], quant_params=quant_params, step_mode=self.step_mode)

        i = 2
        for in_filters, out_filters, kernel, pool_size, padding, stride, dropout in zip(filters,
                                                                               filters[1:],
                                                                               kernel_sizes[1:],
                                                                               pool_sizes[1:],
                                                                               paddings[1:],
                                                                               strides[1:],
                                                                               dropout_list[1:]):
            layers[f'conv{i}'] = sjlayers_t.QuantizedConv(in_channels=in_filters,
                                                          out_channels=out_filters,
                                                          kernel_size=kernel,
                                                          padding=padding,
                                                          stride=stride,
                                                          bias=not batch_norm,
                                                          quant_params=quant_params,
                                                          step_mode=self.step_mode)

            if batch_norm:
                layers[f'bn{i}'] = sjlayers_t.QuantizedBatchNorm(out_filters,
                                                                 quant_params=quant_params,
                                                                 step_mode=self.step_mode)

            layers[f'neuron{i}'] = self.create_neuron(quant_params=quant_params)

            if dropout:
                layers[f'dropout{i}'] = Dropout(dropout, step_mode=self.step_mode)
            if pool_size:
                layers[f'maxpool{i}'] = sjlayers_t.QuantizedMaxPool(pool_size, quant_params=quant_params, step_mode=self.step_mode)

            i += 1

        if (math.prod(postpool) if  isinstance(postpool, list) else postpool) > 1:
            layers['postpool'] = sjlayers_t.QuantizedAvgPool(tuple(postpool) if isinstance(postpool, list) else postpool,
                                                    step_mode=self.step_mode,
                                                    quant_params=quant_params)

        if gsp:
            layers['gsp'] = layers_t.QuantizedGlobalSumPool(quant_params=quant_params)
        else:
            layers['flatten'] = Flatten(step_mode=self.step_mode)

            in_features = np.array(input_shape[:-1]) // np.array(prepool)
            for _, kernel, pool, padding, stride in zip(filters, kernel_sizes, pool_sizes, paddings, strides):
                in_features += np.array(padding) * 2
                in_features -= (kernel - 1)
                in_features = np.ceil(in_features / stride).astype(int)
                if pool:
                    in_features = in_features // pool
            in_features = in_features // np.array(postpool)
            in_features = in_features.prod()
            in_features *= filters[-1]

            j = 1
            for in_units, out_units, dropout in zip((in_features, *fc_units), fc_units, dropout_list[len(filters):]):
                layers[f'fc{j}'] = QuantizedLinear(in_units, out_units, quant_params=quant_params, step_mode=self.step_mode)
                layers[f'neuron{i}'] = self.create_neuron(quant_params=quant_params)
                if dropout:
                    layers[f'dropout{i}'] = Dropout(dropout, step_mode=self.step_mode)
                i += 1
                j += 1

            layers[f'fc{j}'] = QuantizedLinear(fc_units[-1] if len(fc_units) > 0 else in_features,
                                               output_shape[0],
                                               quant_params=quant_params,
                                               step_mode=self.step_mode)
        self.layers = nn.ModuleDict(layers)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward calls each of the SCNN :attr:`layers` sequentially.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input
        for layer in self.layers:
            x = self.layers[layer](x)

        return x



"""Qualia-Plugin-SNN learningframework package contains the SpikingJelly interface to use with the bundled learningmodels."""

from .SpikingJelly import SpikingJelly
from .SpikingJellyMultiStep import SpikingJellyMultiStep
from .SpikingJellyTimeStepsInData import SpikingJellyTimeStepsInData

__all__ = ['SpikingJelly', 'SpikingJellyMultiStep', 'SpikingJellyTimeStepsInData']



"""Contain implementation of a QuantizedLinear layer with support for SpikingJelly ``step_mode``."""

from __future__ import annotations

import sys

from qualia_core.learningmodel.pytorch.layers import quantized_layers
from qualia_core.typing import TYPE_CHECKING
from spikingjelly.activation_based import functional  # type: ignore[import-untyped]
from spikingjelly.activation_based.base import StepModule  # type: ignore[import-untyped]

# We are inside a TYPE_CHECKING block but our custom TYPE_CHECKING constant triggers TCH001-TCH003 so ignore them
if TYPE_CHECKING:
    import torch  # noqa: TCH002
    from qualia_core.learningmodel.pytorch.Quantizer import QuantizationConfig  # noqa: TCH002
    from torch import nn  # noqa: TCH002

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class QuantizedLinear(quantized_layers.QuantizedLinear,
                      StepModule):  # type: ignore[misc]
    """Add SpikingJelly's ``step_mode`` support to Qualia's quantized Linear layer."""

    def __init__(self,  # noqa: PLR0913
                 in_features: int,
                 out_features: int,
                 quant_params: QuantizationConfig,
                 bias: bool = True,  # noqa: FBT001, FBT002
                 activation: nn.Module | None = None,
                 step_mode: str = 's') -> None:
        """Construct :class:`QuantizedLinear`.

        :param in_features: Dimension of input vector
        :param out_features: Dimension of output vector and number of neurons
        :param bias: If ``True``, adds a learnable bias to the output.
        :param quant_params: Dict containing quantization parameters, see
                             :class:`qualia_core.learningmodel.pytorch.Quantizer.Quantizer`
        :param activation: Activation layer to fuse for quantization purposes, ``None`` if unfused or no activation
        :param step_mode: SpikingJelly's ``step_mode``, either ``'s'`` or ``'m'``, see
                          :class:`spikingjelly.activation_based.layer.Linear`
        """
        super().__init__(in_features=in_features,
                         out_features=out_features,
                         bias=bias,
                         quant_params=quant_params,
                         activation=activation)
        self.step_mode = step_mode

    @override
    def extra_repr(self) -> str:
        """Add ``step_mode`` to the ``__repr__`` method.

        :return: String representation of :class:`torch.nn.Linear` with ``step_mode``.
        """
        return super().extra_repr() + f', step_mode={self.step_mode}'

    # Copied to call super().forward(x) of quantized_layers1d.Conv1d instead of torch.nn.Conv1d
    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        """Forward :class:`qualia_core.learningmodel.pytorch.layers.quantized_layers.QuantizedLinear` with ``step_mode`` support.

        :param input: Input tensor
        :return: Output tensor
        """
        x = input
        if self.step_mode == 's':
            x = super().forward(x)

        elif self.step_mode == 'm':
            x = functional.seq_to_ann_forward(x, super().forward)

        return x
